/*
    A Cycle Timer for StrongARM systems with 21285 bridges at 0xe10000,
      that would be NetWinders.
    Copyright (C) 1998   Jim Studt <jim@federated.com>

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
    This library can measure the number of clock cycles between events
    on a Netwinder.  See nwtest.c for an example.
    
    It uses two count down timers in the PCI bridge chip to do this.
    Each of these timers can count down from at most 2^24 and then
    recycle.  Unfortunately 2^24 is no longer a huge number, the counters
    wrap about 4 times / second on the netwinder.  We mitigate this by 
    using two timers with slightly different periods and checking the
    relative phase change, this extends our useful period to something 
    around a 5 hours.  If it takes longer than that I will posit you 
    really didn't need 15ns resolution.  Use a different tool.
*/

/*
  Jim's analysis of how to computer elapsed time by the relative 
  phase of two signals - may explain the math in this module and
  the choice of constants.

  Consider two periodic signals, f1 and f2 of period M1 and M2 where
  M1 > M2 and M2 >> M1 - M2.

    phase1(x) is the phase of f1(x),  (x mod M1)/M1  (ranges 0 to 1)
    phase2(x) is the phase of f2(x),  (x mod M2)/M2  (ranges 0 to 1)

    let T1 = (x mod M1)
    let T2 = (x mod M2)

    phase1(x) = T1/M1
    phase2(x) = T2/M2

    T1 and T2 are our count down timers from the hardware, actually
    T1 is (M1 - hardware timer).

  Now, consider the time from zero but << M2...  (x mod M1)==x

    time from zero = x

    phase1(x) = x/M1
    phase2(x) = x/M2

    phase2(x) - phase1(x) = x/M1 - x/M2

  Now subsitute in our definitions of phase from above...

    T2/M2 - T1/M1 = x/M2 - x/M1  
    T2*M1 - T1*M2 = M1*x - M2*x
    x = (T2*M1 - T1*M2)/(M1-M2)

  ... viola!  we get time from zero in terms of out constants and timers.

  Now for the hard realities.

  1) We have a division in our computation, this bites on ARM.  We will
     choose M1 and M2 so that this is a shift instead of a division.
     Demand: (M1-M2) is a power of two.

  2) Clock jitter.  Despite our best efforts, we just can not sample
     T1 and T2 at the same time.  Turns out we can't even predict with
     absolute precision the time between the sampling of T1 and T2.
     This ruins the calculation.

     The deal with this we choose M1 and M2 so that M1 and M2 are multiples
     of (M1 - M2).  This means that (T1 - T2) mod (M1 - M2) is zero, or 
     rather should be.  If we choose (M1 - M2) > (twice maximal jitter) 
     then we can just fixup T2 to be the 'right' value before performing
     the computation.

  3) Counter wrap.  Our synthesized counter is going to wrap at 
     (M1 * M2)/(M1 - M2)

  4) The ARM counters are limited to 24 bits.  The maximal jitter 
     in the netwinder appears to be 4, but I'll bet there are odd cases
     that can be much higher, so I choose M1 to be 2^24 and M2 to be
     M1 - 256.  That gives maximal jitter of 127 and about 5 hours
     for the counter to roll over at 15ns resolution.
*/

#include <unistd.h>
#include <stdio.h>
#include <strings.h>
#include <stdlib.h>
#include <asm/hardware.h>
#include <fcntl.h>
#include <sys/mman.h>

#include "nwtimer.h"

#define MOD1 0x01000000
#define MOD2 (MOD1-256)

/*
** the lag from reading t2 after t1.  Probably should set this dynamically
** by reading the counters after starting them, but the obvious way
** to code that is infinitely recursive.  fixme.
*/
static int t2lag = 50;

volatile int *record_time_internal(void)
{
    int fd = open("/dev/kmem",O_RDWR);
    volatile int *km = 0;

    /*
    ** we mmap /dev/kmem to get read/write access to our registers.
    ** You will have to be root to do this.
    */
    if ( fd == -1) {
	perror("failed to open /dev/kmem");
	exit(1);
    }
    km = mmap(0, 0x400, PROT_READ|PROT_WRITE, MAP_SHARED, fd, IO_BASE_ARM_CSR);
    if ( km == (int *)-1) {
	perror("failed to mmap /dev/kmem");
	exit(1);
    }

    /*
    ** we will insist that the timers are either zero loaded, which makes
    ** no sense so we figure they are unused, or loaded with out values.
    ** This should keep us from tromping on another use of the timers.
    ** We can only hope that others are as courteous to us.
    */
    if ( km[TIMER_2_LOAD/4] != (MOD1-1) && km[TIMER_3_LOAD/4] != (MOD2-1)) {
        /*
	** already loaded.  assume we are ok (probably should check phase)
	*/
    } else {
        if ( km[TIMER_2_LOAD/4] ||  km[TIMER_3_LOAD/4] && ) {
	    fprintf(stderr, "timer2 and time3 are not available\n");
	    exit(1);
	}

	/*
	** Start the timers, the load values are 256 different, so the counters
	** will diverge by 256 each time they wrap.
	*/
	km[TIMER_2_CONTROL/4] = 0x0000;  /* stop timers */
	km[TIMER_3_CONTROL/4] = 0x0000;
	km[TIMER_2_LOAD/4] = MOD1-1;
	km[TIMER_3_LOAD/4] = MOD2-1;
	km[TIMER_2_CONTROL/4] = 0x00c0;  /* reload when zero, start timers */
	km[TIMER_3_CONTROL/4] = 0x00c0;

	/*
	** should initialize t2lag here
	*/
    }

    return km;
}

static long long f1( const struct nwtimer *t)
{
    int ft2;
    int d;

    /*
    ** ft2 is the 'fudged t2'.  First take out the constant delay that
    ** comes from reading t2 after t1.  (note: we may exceed MOD2 after
    ** this operation, but its ok, we fix that later) 
    */
    ft2 = t->t2 + t2lag;

    /*
    ** now we de-jitter ft2.  The timing between t1 and t2 varies by a
    ** few cycles because of various other bus operations, so we lock
    ** it to the nearest multiple of MOD1-MOD2 here.
    ** This is where when MOD1 and MOD2 are too close together for the
    ** jitter in your hardware you will get nailed.
    */
    d = (t->t1 - ft2) & 0xff;
    if ( d > 128) d -= 256;
    ft2 += d;

    /*
    ** we require that t2's phase  must be ahead of t1's phase because 
    ** it is faster, so bump t2 forward if it is behind.  (MOD2 must be
    ** close to MOD1 for this to work.)
    */
    if ( ft2 < t->t1) ft2 += MOD2;

    /*
    ** this is the real calculation.  See its derivation in the file
    ** commend for more info.
    */
    return ( ((long long)MOD2-ft2)*MOD1 - ((long long)MOD1-t->t1)*MOD2)/
	(MOD1-MOD2);
}
 
unsigned long long cycles_between( const struct nwtimer *start,
				   const struct nwtimer *end)
{
    long long v;

    v = f1(end) - f1(start);
    if ( v < 0) v += ((long long)MOD1*MOD2/(MOD1-MOD2));
    return v;
}

unsigned long long cycles_to( const struct nwtimer *t)
{
    long long v = f1(t);
    if ( v < 0) v += ((long long)MOD1*MOD2/(MOD1-MOD2));
    return v;
}






